#include "geo_next_node.h"

GeoNextNode::GeoNextNode(int algo, nsaddr_t addr, int M, int R, int S, int P, int THR_R, int THR_F, Trace * logtarget, int geo_trace) : x_(0), y_(0), z_(0) {

	algo_ = algo;
	addr_ = addr;

	M_ = M;
	R_ = R;
	S_ = S;
	P_ = P;

	THR_R_ = THR_R;
	THR_F_ = THR_F;

	logtargetnext_ = logtarget;
	geo_trace_ = geo_trace;

	geo_data_ = GEO_data();
}

// Returns the address of the next node
// -1 if the algorithm number is not valid or if the algorithm fails.
// -3 if the target are multiple nodes (flooding or restricted flooding)
nsaddr_t GeoNextNode::nextNode(MobileNode* mn, Packet* data, GNeighbors nbr) {
	trace("geo_data size all'inizio= %d", geo_data_.size());
	// take node's current position
	x_ = (float)mn->X();
	y_ = (float)mn->Y();
	z_ = (float)mn->Z();

	// take data packet's headers
	hdr = HDR_GEO(data);
	ch = HDR_CMN(data);

	// take neighbors object
	nbr_ = nbr;
	
	// Check if the neighborood is empty
	if(nbr_.empty()) {
		trace("[%d] *** ERROR: neighborhood is empty!!! ***\n", addr_);
		return FAIL;
	}

	// Check if any neighbor is the destination node
	for(int i = 0; i < nbr_.size(); i++)
	{	
		if(nbr_.get(i)->addr_ == hdr->dst_) {
			trace("destination found!");
			return nbr_.get(i)->addr_;
		}
	}

	nsaddr_t target;

	switch (algo_)
	{
		case (GREEDY): // GREEDY algorithm
		return progress(GREEDY, false, nbr_);

		case (COMPASS): // COMPASS algorithm
		return progress(COMPASS, false, nbr_);

		case (MFR): // MOST FORWARD algorithm
		return progress(MFR, false, nbr_);

		case (ELLIPSOID): // ELLIPSOID algorithm
		return progress(ELLIPSOID, false, nbr_);

		case (PAB3D): // PAB3D algorithm
		if(ch->last_hop_ < 0) { // it is the sender node
			hdr->ttlr_ = THR_R_;
		}
		if(hdr->ttlr_ == 0) { // time to live expired
			return -1;
		}
		hdr->ttlr_--;
		return rw().front();

		case (PROJECTIVE): // Projective Face algorithm
		if(ch->last_hop_ < 0)
		{
			setSDP();
			hdr->ttlf_ = THR_F_;
			hdr->nFace_ = 1;
			return face2(hdr->x1_, hdr->y1_, hdr->z1_, hdr->x2_, hdr->y2_, hdr->z2_,
			hdr->x3_, hdr->y3_, hdr->z3_, true);
		}
		hdr->ttlf_ --;
		if(hdr->ttlf_ == 0 && hdr->nFace_ == 1) {
			trace("\tChange orthogonal plane");
			setSDO();
			hdr->ttlf_ = THR_F_;
			hdr->nFace_ = 2;
			return face2(hdr->x1_, hdr->y1_, hdr->z1_, hdr->x2_, hdr->y2_, hdr->z2_,
			hdr->x3_, hdr->y3_, hdr->z3_, true);
		} else if(hdr->ttlf_ == 0 && hdr->nFace_ == 2)
			return -1;
		
		return face2(hdr->x1_, hdr->y1_, hdr->z1_, hdr->x2_, hdr->y2_, hdr->z2_,
			hdr->x3_, hdr->y3_, hdr->z3_, false);

		case (CFACE3): // CFACE(3) algorithm
		
			if(ch->last_hop_ < 0)
			{
				trace("\tStart with XY plane");
				hdr->ttlf_ = THR_F_;
				hdr->nFace_ = 1;
				return face2(0,0,0, 1,0,0, 0,1,0, true);
			}

			if(hdr->nFace_ == 1)
			{
				if(hdr->ttlf_ == 0)
				{
					trace("\tRestart with XZ plane");
					hdr->ttlf_ = THR_F_;
					hdr->nFace_ = 2;
					return face2(0,0,0, 1,0,0, 0,0,1, true);
				}
				hdr->ttlf_--;
				return face2(0,0,0, 1,0,0, 0,1,0, false);
			}

			if(hdr->nFace_ == 2)
			{
				if(hdr->ttlf_ == 0)
				{
					trace("\tRestart with YZ plane");
					hdr->ttlf_ = THR_F_;
					hdr->nFace_ = 3;
					return face2(0,0,0, 0,1,0, 0,0,1, true);
				}
				hdr->ttlf_--;
				return face2(0,0,0, 1,0,0, 0,0,1, false);
			}

			if(hdr->nFace_ == 3)
			{
				if(hdr->ttlf_ == 0)
				{
					// fail
					return -1;
				}
				hdr->ttlf_--;
				return face2(0,0,0, 0,1,0, 0,0,1, false);
			}
		

		case(ALSP): // ALSP algorithm
		
			if(ch->last_hop_ < 0)
			{
				setLSP();
				hdr->abs_ = THR_R_;
				hdr->ttlf_ = THR_F_;
				hdr->nFace_ = 1;
				return face2(hdr->x1_, hdr->y1_, hdr->z1_, hdr->x2_, hdr->y2_, hdr->z2_,
				hdr->x3_, hdr->y3_, hdr->z3_, true);	
			}

			if(hdr->ttlf_ == 0) {
				hdr->nFace_++;
				hdr->ttlf_ = THR_F_;
				if(hdr->nFace_ == 2) { trace("\tRenew plane (1)");
					setSDP();
				}
				if(hdr->nFace_ == 3) { trace("\tRenew plane (2)");
					setSDO();
				}
				if(hdr->nFace_ == 4) {
					return -1;
				}
				return face2(hdr->x1_, hdr->y1_, hdr->z1_, hdr->x2_, hdr->y2_, hdr->z2_,
				hdr->x3_, hdr->y3_, hdr->z3_, true);	
			}

			if(hdr->abs_ == 0)
			{ 	
				// refresh LSP
				trace("\tRefresh LSP");
				setLSP();
				hdr->abs_ = THR_R_;
				return face2(hdr->x1_, hdr->y1_, hdr->z1_, hdr->x2_, hdr->y2_, hdr->z2_,
				hdr->x3_, hdr->y3_, hdr->z3_, true);		
			}

			hdr->abs_--;
			hdr->ttlf_--;
			return face2(hdr->x1_, hdr->y1_, hdr->z1_, hdr->x2_, hdr->y2_, hdr->z2_,
				hdr->x3_, hdr->y3_, hdr->z3_, false);
		
		case(GALSPG):  // G-ALSP-G algorithm

			target = -2;
			if(ch->last_hop_ < 0)
			{
				target = progress(GREEDY, false, nbr_);
				hdr->closest_ = addr_;
				hdr->nFace_ = 0;
			} 
			else
			{
				float closeX, closeY, closeZ;
				MobileNode *closestTemp = (MobileNode *) (Node::get_node_by_address(hdr->closest_));
				closeX = closestTemp->X();
				closeY = closestTemp->Y();
				closeZ = closestTemp->Z();
				trace("\tNFace = %d", hdr->nFace_);
				if(hdr->nFace_ == 0 || 
				   dist(x_,     y_,     z_,      hdr->dstX_, hdr->dstY_, hdr->dstZ_) < 
				   dist(closeX, closeY, closeZ , hdr->dstX_, hdr->dstY_, hdr->dstZ_) - 0.01)
				{
				

					trace("\tdist is less: try in GREEDY");

					target = progress(GREEDY, false, nbr_);
					hdr->closest_ = addr_;

					trace("\tTarget for greedy is %d", target );

				}
			}

			if(target == -1) {
				
				// Greedy -> Face
				trace("\tStart with FACE");
				hdr->nFace_ = 1;
				setLSP();
				hdr->ttlf_ = THR_F_;
				hdr->abs_ = THR_R_;
				//return face2(hdr->x1_, hdr->y1_, hdr->z1_, hdr->x2_, hdr->y2_, hdr->z2_,
				//hdr->x3_, hdr->y3_, hdr->z3_, true);	
				return face2(0,0,0, 1,0,0, 0,1,0, true);
			} else if(target == -2) {
				trace("\tContinue with FACE");
				// Face -> Face
				if(hdr->ttlf_ == 0) {
					return -1; // delete it
					/*
					hdr->nFace_++;
					hdr->ttlf_ = THR_F_;
					if(hdr->nFace_ == 2) {
						setSDP();
					}
					if(hdr->nFace_ == 3) {
						setSDO();
					}
					if(hdr->nFace_ == 4) {
						return -1;
					}
					return face2(hdr->x1_, hdr->y1_, hdr->z1_, hdr->x2_, hdr->y2_, hdr->z2_,
					hdr->x3_, hdr->y3_, hdr->z3_, true);	
					*/
				}
				/*
				if(hdr->abs_ == 0)
				{ // refresh LSP
					trace("\tRenew LSP");
					setLSP();
					hdr->abs_ = THR_R_;
					return face2(hdr->x1_, hdr->y1_, hdr->z1_, hdr->x2_, hdr->y2_, hdr->z2_,
					hdr->x3_, hdr->y3_, hdr->z3_, true);		
				}
				*/

				//hdr->abs_--;
				hdr->ttlf_--;
				//return face2(hdr->x1_, hdr->y1_, hdr->z1_, hdr->x2_, hdr->y2_, hdr->z2_,
				//	hdr->x3_, hdr->y3_, hdr->z3_, false);
				return face2(0,0,0, 1,0,0, 0,1,0, false);
			} else {
				// Greedy -> Greedy | Face -> Greedy
				return target;
			}
			break;

		case(PFACEP): // PAB3D - CFace(1) - PAB3D algorithm
		
		// time to live of face is set as TTLR
		//nsaddr_t target = -1;
		if(ch->last_hop_ < 0)
		{
			hdr->ttlr_ = THR_R_;
			hdr->ttlf_ = 0;
			hdr->nFace_ = 0;
			return rw().front();
		}

		if(hdr->nFace_ == 0) {
			// rw phase
			if(hdr->ttlr_ == 0) {
				// starts face phase
				hdr->nFace_ = 1;
				hdr->ttlf_ = THR_F_;
				int b [4] = {0,0,0,0};
				srand(time(0)+rand());
				int c1=rand()%2;
				int c2=rand()%2;
				b[c1] = 1;
				b[c2+2] = 1;
				if(b[1]==1) b[2]=0;
				if(b[2]==1) b[1]=0;
				if(b[0]==0) b[3]=1;
				if(b[3]==0) b[0]=1;
				hdr->x1_ = 0; hdr->y1_ = 0; hdr->z1_ = 0;
				hdr->x2_ = b[0]; hdr->y2_ = b[1]; hdr->z2_ = 0;
				hdr->x3_ = 0; hdr->y3_ = b[2]; hdr->z3_ = b[3];
				return face2(hdr->x1_, hdr->y1_, hdr->z1_, hdr->x2_, hdr->y2_, hdr->z2_, hdr->x3_, hdr->y3_, hdr->z3_, true);
			}
			hdr->ttlr_--;
			return rw().front();
		}
		else {
			// face phase
			if(hdr->ttlf_ == 0) {
				// start with rw phase
				hdr->nFace_ = 0;
				hdr->ttlr_ = THR_R_;
				return rw().front();
			}
			hdr->ttlf_--;
			return face2(hdr->x1_, hdr->y1_, hdr->z1_, hdr->x2_, hdr->y2_, hdr->z2_, hdr->x3_, hdr->y3_, hdr->z3_, false);
		}

		case(PFACE3):	// PAB3D - CFace(3)
		if(ch->last_hop_ < 0)
		{
			hdr->ttlr_ = THR_R_;
			hdr->nFace_ = 0;
			return rw().front();
		}
		
		if(hdr->nFace_ == 0) {
			if(hdr->ttlr_ > 0)
			{
				hdr->ttlr_--;
				return rw().front();
			}
			else
			{
				hdr->ttlf_ = THR_F_;
				hdr->nFace_ = 1;
				return face2(0,0,0, 1,0,0, 0,1,0, true);
			}
		}

		if(hdr->nFace_ == 1)
		{
			if(hdr->ttlf_ == 0)
			{
				trace("\tStep in plane XZ");
				hdr->ttlf_ = THR_F_;
				hdr->nFace_ = 2;
				return face2(0,0,0, 1,0,0, 0,0,1, true);
			}
			hdr->ttlf_--;
			return face2(0,0,0, 1,0,0, 0,1,0, false);
		}

		if(hdr->nFace_ == 2)
		{
			if(hdr->ttlf_ == 0)
			{
				trace("\tStep in plane YZ");
				hdr->ttlf_ = THR_F_;
				hdr->nFace_ = 3;
				return face2(0,0,0, 0,1,0, 0,0,1, true);
			}
			hdr->ttlf_--;
			return face2(0,0,0, 1,0,0, 0,0,1, false);
		}

		if(hdr->nFace_ == 3)
		{
			if(hdr->ttlf_ == 0)
			{
				trace("\tCFace(3) Failed");
				return -1;
			}
			hdr->ttlf_--;
			return face2(0,0,0, 0,1,0, 0,0,1, false);
		}

		case(GPG):	// GRG (GREEDY-PAB3D-GREEDY)
		if(ch->last_hop_ < 0 || hdr->mode_ == PROGRESS_MODE)
		{
			hdr->closest_ = addr_;
			hdr->mode_ = PROGRESS_MODE;
			target = progress(GREEDY, false, nbr_);
			trace("\tTrying Greedy: target is: %d", target );
		} 
		else // RANDOM_MODE
		{
			float closeX, closeY, closeZ;
			MobileNode *closestTemp = (MobileNode *) (Node::get_node_by_address(hdr->closest_));
			closeX = closestTemp->X();
			closeY = closestTemp->Y();
			closeZ = closestTemp->Z();
			if(dist(x_,     y_,     z_,      hdr->dstX_, hdr->dstY_, hdr->dstZ_) + 0.001 < 
			   dist(closeX, closeY, closeZ , hdr->dstX_, hdr->dstY_, hdr->dstZ_))
			{
			    hdr->closest_ = addr_;
			    hdr->mode_ = PROGRESS_MODE;
				target = progress(GREEDY, false, nbr_);
				trace("\tDist is less, try Greedy: target is: %d", target );
			}
		}

		if(hdr->mode_ == PROGRESS_MODE) {
			if(target == -1) {
				hdr->mode_ = RANDOM_MODE;
				hdr->ttlr_ = THR_R_;
				return rw().front();
			}
			else
				return target;
		}
		else { // RANDOM_MODE
			if(hdr->ttlr_ == 0) return -1;
			hdr->ttlr_ --;
			return rw().front();
		}	

		case(LAR): // LAR algorithm
		hdr->fSize_ = 0;
		
		for(int i = 0; i < nbr_.size() && hdr->fSize_<10; i++) {
			GNode * current = nbr_.get(i);
			
			// check if node is inside the LAR zone (box) or
			// if the node is the previous one
			if(insideBox(current->x_, current->y_, current->z_, 
			hdr->srcX_, hdr->srcY_, hdr->srcZ_, hdr->dstX_, hdr->dstY_, hdr->dstZ_, LAR_R)) {
				hdr->targets_[hdr->fSize_] = current->addr_;
				hdr->fSize_++;
			}
		}
		if(hdr->fSize_ > 0)
			return -3;
		return -1;
		

		case(PLAR): // PAB3D-LAR algorithm
		{
		hdr->fSize_ = 0;
		std::vector<nsaddr_t> targetss = rw();
		for(unsigned int i = 0; i < targetss.size() && hdr->fSize_<10; i++) {
			GNode * current = nbr_.getByAddr(targetss.at(i));
			
			// check if node is inside the LAR zone (box) or
			// if the node is the previous one
			if(insideBox(current->x_, current->y_, current->z_, 
			hdr->srcX_, hdr->srcY_, hdr->srcZ_, hdr->dstX_, hdr->dstY_, hdr->dstZ_, LAR_R)) {
				hdr->targets_[hdr->fSize_] = current->addr_;
				hdr->fSize_++;
			}
		}
		if(hdr->fSize_ > 0)
			return -3;
		return -1;
		} break;
		
		case(DFS): // DEPTH FIRST SEARCH algorithm
		return dfs();
		

		default:
		fprintf(stderr, "*** ERROR: number of protocol (%d) ****", algo_);
		exit(1);
	}
}

// Returns the address of the node that makes a progress toward the destination
// "type" is the progress algorithm, 
// "back" indicates whether permit (true) or not (false) a return to the previous node.
nsaddr_t GeoNextNode::progress(int type, bool back, GNeighbors nbr) {

	nsaddr_t target = -1;
	if(nbr.empty())
		return target;
	//trace("dest coord: %f, %f, %f\n", hdr->dstX_, hdr->dstY_, hdr->dstZ_);
	if(type == CASUAL) return nbr.get(rand() % nbr.size())->addr_;
	float min;
	if(type == GREEDY) {
		if(back) min = dist(hdr->dstX_, hdr->dstY_, hdr->dstZ_, nbr.front()->x_, nbr.front()->y_, nbr.front()->z_);
		else min = dist(hdr->dstX_, hdr->dstY_, hdr->dstZ_, x_, y_, z_);
	}
	else if(type == COMPASS) min = angle(x_, y_, z_, hdr->dstX_, hdr->dstY_, hdr->dstZ_, nbr.front()->x_, nbr.front()->y_, nbr.front()->z_);
	else if(type == MFR) min = projDist(x_, y_, z_, hdr->dstX_, hdr->dstY_, hdr->dstZ_, nbr.front()->x_, nbr.front()->y_, nbr.front()->z_);
	else if(type == ELLIPSOID) min = dist(x_, y_, z_, nbr.front()->x_, nbr.front()->y_, nbr.front()->z_) + dist(hdr->dstX_, hdr->dstY_, hdr->dstZ_, nbr.front()->x_, nbr.front()->y_, nbr.front()->z_);
	else { fprintf(stderr, "*** ERROR: number of progress algorithm!! ***"); exit(1);}
	if(back || type==CASUAL || type==COMPASS|| type==MFR || type==ELLIPSOID) target = nbr.front()->addr_;
	float curr;
	for(int i = 0; i < nbr.size(); i++) {
		//trace("current coord (%d): %f, %f, %f\n", nbr.get(i)->addr_, nbr.get(i)->x_, nbr.get(i)->y_, nbr.get(i)->z_);
		if(type == GREEDY) curr = dist(hdr->dstX_, hdr->dstY_, hdr->dstZ_, nbr.get(i)->x_, nbr.get(i)->y_, nbr.get(i)->z_);
		else if(type == COMPASS) curr = angle(x_, y_, z_, hdr->dstX_, hdr->dstY_, hdr->dstZ_, nbr.get(i)->x_, nbr.get(i)->y_, nbr.get(i)->z_);
		else if(type == MFR) curr = projDist(x_, y_, z_, hdr->dstX_, hdr->dstY_, hdr->dstZ_, nbr.get(i)->x_, nbr.get(i)->y_, nbr.get(i)->z_);
		else if(type == ELLIPSOID) curr = dist(x_, y_, z_, nbr.get(i)->x_, nbr.get(i)->y_, nbr.get(i)->z_) + dist(hdr->dstX_, hdr->dstY_, hdr->dstZ_, nbr.get(i)->x_, nbr.get(i)->y_, nbr.get(i)->z_);
		if(curr < min) {
			target = nbr.get(i)->addr_;
			min = curr;
		}
	}
	// fails if returns the previous node (if back is false)
	if(!back && target == ch->last_hop_) 
		return -1;
	
	return target;
}

// Returns the address of the node using the PAB3D algorithm.
// (for single path algorithms, tha target node is on the first position)
// M is the number of candidate nodes
// R determines how to select the candidate neighbors: 1 (GREEDY), 2 (COMPASS), 3 (MFR)
// S determines how to set the next node selection probability: 1 (GREEDY), 2 (COMPASS), 3 (MFR)
// P determines if consider or not the plane/s to divide the choosing regions.
std::vector<nsaddr_t> GeoNextNode::rw() 
{
	std::vector<nsaddr_t> targets;
	
	// work on nb neighborood
	GNeighbors nb(nbr_);

	// Check if the parameter combination is not valid
	if(M_ == 0 || (P_ == 1 && (M_%2) == 0) )
	{
		fprintf(stderr, "*** ERROR: Parameter m or ab are not valid (m = %d, ab = %d) ***", M_, P_);
		exit(1);
	}

	nsaddr_t nbb[M_]; // array of best neighbors choosed
	float prob[M_]; // probability values of best neighbors
	float tot = 0;
	int size = 0; // current number of best neighbors
	nbb[size] = -1; // fist element is null
	prob[size] = 0;
	size++;

	if(nb.size() == 1) {
		// the only chosen node is the prevoius one
		targets.push_back(nb.front()->addr_);
		return targets;
	} else {
		// delete the previous node from the neighborood
		nb.delByAddr(ch->last_hop_);
	}

	// SELECT CANDIDATE NODES
	if(P_ == 0) // no plane
	{
		GNeighbors nbcopy(nb); // temp vector

		for(int im = 0; im < M_ && nbcopy.size() > 0; im++)
		{
			nsaddr_t best;
			best = progress(R_, true, nbcopy);
			nbb[size] = best;
			size++;
			// delete best in nbcopy
			nbcopy.delByAddr(best);

		}
	}
	else // plane
	{
		GNeighbors nbN; // neighbor up
		GNeighbors nbS; // neighbor down
		GNeighbors nbE; // neighbor up
		GNeighbors nbO; // neighbor down
	
		// choose first node n1 with Greedy method
		nsaddr_t n1 = progress(1, true, nb);
		nbb[size] = n1; size++;

		// take n1 coordinates
		float nx, ny, nz;
		GNode * node1 = nb.getByAddr(n1);
		nx = node1->x_; ny = node1->y_; nz = node1->z_;

		// calculate ortogonal coordinate of n1 to line sd
		float nox, noy, noz;
		gpoint ortoN = ortoPoint(x_, y_, z_, hdr->dstX_, hdr->dstY_, hdr->dstZ_, nx, ny, nz);
		nox = ortoN.x;
		noy = ortoN.y;
		noz = ortoN.z;

		// compute plane coefficients a, b, c
		float a1 = ((hdr->dstY_-y_) * (nz-z_)) - ((hdr->dstZ_-z_) * (ny-y_));
		float b1 = - (  ((hdr->dstX_-x_) * (nz-z_)) - ((hdr->dstZ_-z_) * (nx-x_)) );
		float c1 = ((hdr->dstX_-x_) * (ny-y_)) - ((hdr->dstY_-y_) * (nx-x_));
		//printf("a = %f, b = %f, c=%f", a,b,c);
		float a2 = ((hdr->dstY_-y_) * (noz-z_)) - ((hdr->dstZ_-z_) * (noy-y_));
		float b2 = - (  ((hdr->dstX_-x_) * (noz-z_)) - ((hdr->dstZ_-z_) * (nox-x_)) );
		float c2 = ((hdr->dstX_-x_) * (noy-y_)) - ((hdr->dstY_-y_) * (nox-x_));

		for(int i = 0; i < nb.size(); i++)
		{
			float coef1 = a1*(nb.get(i)->x_ - x_) + b1*(nb.get(i)->y_ - y_) + c1*(nb.get(i)->z_ - z_);
			float coef2 = a2*(nb.get(i)->x_ - x_) + b2*(nb.get(i)->y_ - y_) + c2*(nb.get(i)->z_ - z_);
			GNode *node_ = new GNode(nb.get(i)->addr_, nb.get(i)->x_, nb.get(i)->y_, nb.get(i)->z_, nb.get(i)->ts_, NULL);
			if(coef1 < 0) {
				if(M_ == 5 && coef2 < 0)
					nbO.addNode(node_);
				else
					nbN.addNode(node_);
			}
			else if(coef1 > 0) {
				if(M_ == 5 && coef2 < 0)
					nbS.addNode(node_);
				else
					nbE.addNode(node_);
			}
		}
		
		nsaddr_t n2 = progress(R_, true, nb);
		if(n2 != -1 && n2 != n1) {nbb[size] = n2; size++;}
		nsaddr_t n3 = progress(R_, true, nb);
		if(n3 != -1 && n3 != n1 && n3 != n2) {nbb[size] = n3; size++;}
		//printf("n1 = %d, n2 = %d, n3 = %d",  n1,n2,n3);
		if(M_ == 5)
		{
			nsaddr_t n4 = progress(R_, true, nb);
			if(n4 != -1 && n4 != n1 && n4 != n2 && n4 != n3) {nbb[size] = n4; size++;}
			nsaddr_t n5 = progress(R_, true, nb);
			if(n5 != -1 && n5 != n1 && n5 != n2 && n5 != n3 && n5 != n4) {nbb[size] = n5; size++;}
		}
	}

	// Trace the list of candidates
	char cand[255];
	cand[0]=0;
	for(int i =1; i < size; i++) {
		char ad[5];
		sprintf(ad, "%d ", nbb[i]);
		strcat(cand, ad);
	}
	trace("\tCandidate nodes: %s", cand);

	// only one node found
	if(size == 2) {
		targets.push_back(nbb[1]);
		return targets;
	}

	// Partial Flooding algorithms
	if(hdr->mode_ == FLOOD_MODE) {
		for(int i = 1; i < size; i++)
			targets.push_back(nbb[i]);
		return targets;
	}

	// CALCULATE PROBABILITY OF EACH NODE
	for(int i = 1; i < size; i++)
	{
		for(int j = 0; j < nb.size(); j++)
			if(nb.get(j)->addr_ == nbb[i])
			{
				if(S_ == GREEDY)
					tot += dist(nb.get(j)->x_, nb.get(j)->y_, nb.get(j)->z_, hdr->dstX_, hdr->dstY_, hdr->dstZ_);
				else if(S_ == COMPASS)
					tot += angle(x_, y_, z_, hdr->dstX_, hdr->dstY_, hdr->dstZ_, nb.get(j)->x_, nb.get(j)->y_, nb.get(j)->z_);
				else if(S_ == MFR)
					tot += projDist(x_, y_, z_, hdr->dstX_, hdr->dstY_, hdr->dstZ_, nb.get(j)->x_, nb.get(j)->y_, nb.get(j)->z_);
			}
	}

	for(int i = 1; i < size; i++)
	{
		for(int j = 0; j < nb.size(); j++)
			if(nb.get(j)->addr_ == nbb[i])
			{
				if(S_ == 1)
					prob[i] = prob[i-1] + 1-(dist(nb.get(j)->x_, nb.get(j)->y_, nb.get(j)->z_, hdr->dstX_, hdr->dstY_, hdr->dstZ_) / tot);
				else if(S_ == 2)
					prob[i] = prob[i-1] + 1-(angle(x_, y_, z_, hdr->dstX_, hdr->dstY_, hdr->dstZ_, nb.get(j)->x_, nb.get(j)->y_, nb.get(j)->z_) / tot);
				else if(S_ == 3)
					prob[i] = prob[i-1] + 1-(projDist(x_, y_, z_, hdr->dstX_, hdr->dstY_, hdr->dstZ_, nb.get(j)->x_, nb.get(j)->y_, nb.get(j)->z_) / tot);
				else  // S == 0
					prob[i] = prob[i-1] + 1;
			}
	}
	// Trace the probabilities of candidates
	char probt[255];
	probt[0]=0;
	for(int i =1; i < size; i++) {
		char ad[5];
		sprintf(ad, "%.3f ", prob[i]);
		strcat(probt, ad);
	}
	trace("\tProbabilities: %s", probt);

	// CHOOSE NEXT NODE
	srand(time(NULL) + rand());
	float choose = static_cast <float> (rand()) / (static_cast <float> (RAND_MAX/prob[size-1]));
	for(int i = 1; i < size; i++)
	{	
		if(choose > prob[i-1] && choose <= prob[i])
		{
			targets.push_back(nbb[i]);
			return targets;
		}
	}
	targets.push_back(-1);
	return targets;
}

void GeoNextNode::createProj(float x1, float y1, float z1, 
					   float x2, float y2, float z2,
					   float x3, float y3, float z3, 
					   GNeighbors *nbrProj) {
	nbrProj->clear();
	for(int i = 0; i < nbr_.size(); i++) {
		nbrProj->addNode(nbr_.get(i)->project(x1,y1,z1,x2,y2,z2,x3,y3,z3));
	}

}

nsaddr_t GeoNextNode::findCloserOnFace(GNeighbors nbrg, nsaddr_t prev, float cx, float cy, float px, float py)
{
	nsaddr_t choose = prev;
	float baseTg = tg(cx, cy, px, py);
	float minDiffTg = 3*M_PI;
	float diffTg;
	for(int i = 0; i < nbrg.size(); i++) {
		diffTg = tg(cx, cy, nbrg.get(i)->x_, nbrg.get(i)->y_) - baseTg;
		if(nbrg.get(i)->addr_ == prev && diffTg == 0) continue;
		if (diffTg < 0)
      		diffTg += 2*M_PI;
    	if (diffTg < 0)
      		diffTg += 2*M_PI;
      	trace("with node %d -> diffTG = %f",nbrg.get(i)->addr_, diffTg);
      	if(diffTg < minDiffTg) {
      		minDiffTg = diffTg;
      		choose = nbrg.get(i)->addr_;
      	}
	}
	trace("inside findCloserOnFace %d", choose);
	return choose;
}


void GeoNextNode::createGG(GNeighbors nb, GNeighbors *nbrGG) {
	
	// Creo il grafo planare aggiornando nbrGG
	for(int i = 0; i < nb.size(); i++) {	
		float px = (x_ + nb.get(i)->x_) / 2;
		float py = (y_ + nb.get(i)->y_) / 2;
		float pz = (z_ + nb.get(i)->z_) / 2;
		float radp = dist(x_, y_, z_, nb.get(i)->x_, nb.get(i)->y_, nb.get(i)->z_) / 2;

		for(int j = 0; j < nb.size(); j++) {
			float distm = dist(px, py, pz, nb.get(j)->x_, nb.get(j)->y_, nb.get(j)->z_);
			if(distm <= radp && i != j)
			{
				nbrGG->delByAddr(nb.get(j)->addr_);
				break;
			}
		}
	}
}

void GeoNextNode::setLSP() { // set Least Square Plane
	float sx, sy, sz, sxx, syy, sxy, sxz, syz, mx, my, mz;
	sx=x_+hdr->dstX_;
	sy=y_+hdr->dstY_;
	sz=z_+hdr->dstZ_;
	sxx=(x_*x_)+(hdr->dstX_*hdr->dstX_);
	syy=(y_*y_)+(hdr->dstY_*hdr->dstY_); 
	sxy=(x_*y_)+(hdr->dstX_*hdr->dstY_);
	sxz=(x_*z_)+(hdr->dstX_*hdr->dstZ_);
	syz=(y_*z_)+(hdr->dstY_*hdr->dstZ_);
	mx=0; my=0; mz=0;

	for(int i = 0; i < nbr_.size(); i++) {
		sx += nbr_.get(i)->x_;
		sy += nbr_.get(i)->y_;
		sz += nbr_.get(i)->z_;
		sxx += nbr_.get(i)->x_ * nbr_.get(i)->x_;
		syy += nbr_.get(i)->y_ * nbr_.get(i)->y_;
		sxy += nbr_.get(i)->x_ * nbr_.get(i)->y_;
		sxz += nbr_.get(i)->x_ * nbr_.get(i)->z_;
		syz += nbr_.get(i)->y_ * nbr_.get(i)->z_;
		mx = sx / (nbr_.size()+2);
		my = sy / (nbr_.size()+2);
		mz = sz / (nbr_.size()+2);
	}

	float A[3][3] = { {sxx, sxy, sx}, {sxy, syy, sy}, {sx, sy, (float)nbr_.size()+2}};
	float B [3] = {sxz, syz, sz};
	float v[3];
	solve(3, A, B, v);

	//fprintf(stderr, "a b c = (%f, %f, %f), ", x[0], x[1], x[2]);
	hdr->x1_ = mx;
	hdr->y1_ = my;
	hdr->z1_ = mz;
	hdr->x2_ = hdr->srcX_;
	hdr->y2_ = hdr->srcY_;
	hdr->z2_ = hdr->srcX_*v[0] + hdr->srcY_*v[1] + v[2];
	hdr->x3_ = hdr->dstX_;
	hdr->y3_ = hdr->dstY_;
	hdr->z3_ = hdr->dstX_*v[0] + hdr->dstY_*v[1] + v[2];
}

void GeoNextNode::setSDP() { // set Random plane from S to D
	float cx, cy, cz;
	srand(time(NULL));
	cx = rand() % 100 - 50;
	cy = rand() % 100 - 50;
	cz = rand() % 100 - 50;
	hdr->x1_ = hdr->srcX_;
	hdr->y1_ = hdr->srcY_;
	hdr->z1_ = hdr->srcZ_;
	hdr->x2_ = hdr->dstX_;
	hdr->y2_ = hdr->dstY_;
	hdr->z2_ = hdr->dstZ_;
	hdr->x3_ = hdr->x1_ + cx;
	hdr->y3_ = hdr->y1_ + cy;
	hdr->z3_ = hdr->z1_ + cz;
}

void GeoNextNode::setSDO() {
	gpoint orto;
	orto = ortoPoint(hdr->x1_, hdr->y1_, hdr->z1_,hdr->x2_, hdr->y2_, hdr->z2_,hdr->x3_, hdr->y3_, hdr->z3_);
	hdr->x3_ = orto.x;
	hdr->y3_ = orto.y;
	hdr->z3_ = orto.z;
}

void GeoNextNode::setNextP() {}


nsaddr_t GeoNextNode::face2(float x1, float y1, float z1, 
					float x2, float y2, float z2,
					float x3, float y3, float z3, bool start) {

	//trace("Calling Face 2");
	//check errors in 3 point coordinates
	if((x1==x2&&y1==y2&&z1==z2) || (x1==x3&&y1==y3&&z1==z3) || (x3==x2&&y3==y2&&z3==z2))
	{
		fprintf(stderr, "*** ERROR: (Calling Face routing) the 3 plane points are complanar or overlap! ***");
		exit(1);
	}

	nsaddr_t target;
	
	// coordinate proiett del nodo corrente
	GNode *thisNode = new GNode(addr_, x_, y_, z_, 0, NULL);
	GNode *thisProj = thisNode->project(x1,y1,z1,x2,y2,z2,x3,y3,z3);
	float xp = thisProj->x_;
	float yp = thisProj->y_;

	trace("start point = (%f, %f)", xp, yp);
	

	GNode *desti = new GNode(hdr->dst_, hdr->dstX_, hdr->dstY_, hdr->dstZ_, 0, NULL);
	GNode *dstProj = desti->project(x1,y1,z1,x2,y2,z2,x3,y3,z3);

	
	
	GNeighbors nbrP;
	createProj(x1,y1,z1,x2,y2,z2,x3,y3,z3, &nbrP);

	// solo per grafi 2D
	//std::vector<GNode*> nbrP;
	//for(unsigned i = 0; i < nbr_.size(); i++) 
	//	nbrP.push_back(new GNode(nbr_.at(i)->addr_, nbr_.at(i)->x_, nbr_.at(i)->y_ ,nbr_.at(i)->z_));
	createGG(nbr_, &nbrP);
	

	//GNeighbors nbrP;
	//createProj(x1,y1,z1,x2,y2,z2,x3,y3,z3, &nbrP);


	/*
	fprintf(stderr, "Vicini di %d Prima:", addr_);
	for(unsigned i = 0; i < nbr_.size(); i++)
		fprintf(stderr, "%d (%f, %f)", nbr_.at(i)->addr_, nbr_.at(i)->x_, nbr_.at(i)->y_);
	fprintf(stderr,"");
	fprintf(stderr, "Dopo:");
	for(unsigned i = 0; i < nbrP.size(); i++)
		fprintf(stderr, "%d (%f, %f)", nbrP.at(i)->addr_, nbrP.at(i)->x_, nbrP.at(i)->y_);
	fprintf(stderr,"");
	*/
	

	// se questo nodo è quello iniziale del face routing
	if(start || ch->last_hop_ < 0)
	{
		hdr->firstX_ = xp;
		hdr->firstY_ = yp;
		trace("start face: id_prev %d, (%f, %f)---(%f, %f)", ch->last_hop_, xp, yp, dstProj->x_, dstProj->y_);
		return findCloserOnFace(nbrP, ch->last_hop_, xp, yp, dstProj->x_, dstProj->y_);
	}

	// prev projected node
	MobileNode *prevNode = (MobileNode *) (Node::get_node_by_address(ch->last_hop_));

	GNode *prev = new GNode(ch->last_hop_, prevNode->X(), prevNode->Y(), prevNode->Z(), 0, NULL);
	GNode *prevProj = prev->project(x1,y1,z1,x2,y2,z2,x3,y3,z3);

	target = findCloserOnFace(nbrP, ch->last_hop_, xp, yp, prevProj->x_, prevProj->y_);
	trace("continua face: id_prev %d, (%f, %f)---(%f, %f)", ch->last_hop_, xp, yp, prevProj->x_, prevProj->y_);

	// next projected node
	MobileNode *nextNode = (MobileNode *) (Node::get_node_by_address(target));
	GNode *next = new GNode(target, nextNode->X(), nextNode->Y(), nextNode->Z(), 0, NULL);
	GNode *nextProj = next->project(x1,y1,z1,x2,y2,z2,x3,y3,z3);

	int a = 0;
	while(intersect(hdr->firstX_, hdr->firstY_, dstProj->x_, dstProj->y_,
		xp, yp, nextProj->x_, nextProj->y_))
	{	a++;
		trace("intersection with node %d", target);
		if(a > 20){ fprintf(stderr, "ANDATO IN LOOP"); return -1;}
		//fprintf(stderr, "intersect tra %d e %d ",addr_, next->addr_);
		//fprintf(stderr, "(%f, %f)-(%f, %f) and (%f, %f)-(%f, %f)", hdr->firstX_, hdr->firstY_, dstProj->x_, dstProj->y_,
		//xp, yp, nextProj->x_, nextProj->y_);
		float intersX, intersY;
		get_line_intersection(hdr->firstX_, hdr->firstY_, dstProj->x_, dstProj->y_,
		xp, yp, nextProj->x_, nextProj->y_, &intersX, &intersY);
		hdr->firstX_ = intersX;  // accorcia la linea sd
		hdr->firstY_ = intersY;
		//fprintf(stderr, "intx = %f, Inty = %f",intersX, intersY);
		if(orientation(intersX, intersY, dstProj->x_, dstProj->y_, xp, yp) == 1)
			return target;

		target = findCloserOnFace(nbrP, target, xp, yp, intersX, intersY);
		nextNode = (MobileNode *) (Node::get_node_by_address(target));
		next = new GNode(target, nextNode->X(), nextNode->Y(), nextNode->Z(), 0, NULL);
		nextProj = next->project(x1,y1,z1,x2,y2,z2,x3,y3,z3);
	}
	return target;
}

nsaddr_t GeoNextNode::dfs() {

trace("geo_data size = %d", geo_data_.size());
	DFS_record * curr_dfs = geo_data_.dfs_get(ch->uid());

	// it's a new packet
	if(curr_dfs == NULL) {
		trace("[%d] new packet %d in this node", addr_, ch->uid());
		if(geo_data_.size() >= 500)
			fprintf(stderr, "SPACE FULL!! \n");
	
		geo_data_.dfs_add(ch->uid(), ch->last_hop_, -1, nbr_);
		//trace("size = %d", geo_data_.size());
		curr_dfs = geo_data_.dfs_get(ch->uid());

		trace("Candidates:");
		char cand[255];
		cand[0]=0;
		curr_dfs->keep_on.print(cand);
		trace("%s",cand);



		curr_dfs->deleted.push_back(curr_dfs->keep_on.delByAddr(ch->last_hop_));
	} // proceed with next node 
	else if(!hdr->returned_) {
		trace("[%d] Resend back istantly!", addr_);
		hdr->returned_ = true;
		curr_dfs->deleted.push_back(curr_dfs->keep_on.delByAddr(ch->last_hop_));
		return ch->last_hop_;
	} 

	// it's a returned packet
	if(hdr->returned_) {
		trace("[%d] returned packet", addr_);
		curr_dfs->deleted.push_back(curr_dfs->keep_on.delByAddr(ch->last_hop_));
	} // proceed with next node

	if(curr_dfs->size() == 0) {
		trace("[%d] No neighbors, return to from [%d]", addr_, curr_dfs->from);
		hdr->returned_ = true;
		geo_data_.dfs_del(ch->uid());
		trace("[%d] Remove packet from store: size = %d -> %d", addr_, geo_data_.size()+1, geo_data_.size());
		if(addr_ == hdr->src_) {
			trace("addr = %d, src = %d\n", addr_, hdr->src_);
			return -1;
		}
		return curr_dfs->from;
	}

	// checks for new neighbor nodes, and add these in keep_on list
	for(int i = 0; i < nbr_.size(); i++) {
		bool found = false;
		GNode * nn = nbr_.get(i);
		for(int j = 0; j < curr_dfs->deleted.size() && !found; j++) {
			if(nbr_.get(i)->addr_ == curr_dfs->deleted.at(j))
				found = true;
		}
		
		for(int j = 0; j < curr_dfs->keep_on.size() && !found; j++) {
			if(nbr_.get(i)->addr_ == curr_dfs->keep_on.get(j)->addr_)
				found = true;
		}

		if(!found) {
			trace("add new node");
			curr_dfs->keep_on.addNode(nn);
		}
		
	}

	God* god = God::instance();
	trace("[%d] Calculate next active node", addr_);
	nsaddr_t target = progress(GREEDY, true, curr_dfs->keep_on);
	// proceed with next node
	int ii=10;
	while(god->hops(addr_, target) > 1 && ii > 0) {
		
		trace("Target %d is not yet neighbor! current dist = %d", target, god->hops(addr_, target));
		curr_dfs->deleted.push_back(curr_dfs->keep_on.delByAddr(target));

		char cand[255];
		cand[0]=0;
		curr_dfs->keep_on.print(cand);
		trace("actual size: %d: %s",curr_dfs->keep_on.size(),cand);

		target = progress(GREEDY, true, curr_dfs->keep_on);

		if(curr_dfs->keep_on.size() == 0) return -1;
		ii--;
	}
	curr_dfs->active = target;
	hdr->returned_ = false;

	trace("DFS record: from=%d, active=%d, target=%d", curr_dfs->from, curr_dfs->active, target);
	//fprintf(stderr,"DFS record: from=%d, active=%d\n", curr_dfs->from, curr_dfs->active);
	trace("geo_data size = %d", geo_data_.size());
	return target;
}

void
GeoNextNode::trace(char* fmt, ...)
{
	if(geo_trace_) {
	  va_list ap;
	  
	  if (!logtargetnext_) return;

	  va_start(ap, fmt);
	  vsprintf(logtargetnext_->pt_->buffer(), fmt, ap);
	  logtargetnext_->pt_->dump();
	  va_end(ap);
	}
}

void GEO_data::dfs_add(int id, nsaddr_t from, nsaddr_t active, GNeighbors nbr) {
DFS_record * newDFS = new DFS_record(id, from, active, nbr, this);
 	dfs_data_[dfs_size] = newDFS;
 	dfs_data_[dfs_size]->dfs_timer_.force_cancel();
 	dfs_data_[dfs_size]->dfs_timer_.resched(30);
 	dfs_size++;
 }

void GEO_data::dfs_del(int p_id) {
	 	for(int i = 0; i < dfs_size; i++) {
	 		if(dfs_data_[i]->id == p_id && i != dfs_size-1) {
	 			for(int j = i; j < dfs_size-1; j++) {
	 				dfs_data_[j] = dfs_data_[j+1];
	 			}
	 		}
	 		dfs_size--;
	 		return;
	 	}
	 }


DFS_record * GEO_data::dfs_get(int p_id) {
	 	for(int i = 0; i < dfs_size; i++)
	 		if(dfs_data_[i]->id == p_id)
	 			return  dfs_data_[i];
	 	return NULL;
	 }

DFS_record::DFS_record(int id, nsaddr_t from, nsaddr_t active, GNeighbors nbr, GEO_data * gd) : id(id), from(from), active(active), keep_on(nbr), dfs_timer_(id, gd) {

}

GEO_data::GEO_data() {
	 	 dfs_size = 0;
	 	 //dfs_data_ = new DFS_record[200];
	 }

void DFSAliveTimer::expire(Event *e) {
	geo_data_->dfs_del(id_);
}