#include <vector>
#include <stdlib.h>
#include <cmath>
#include <random.h>
#include <trace.h>

/*
static double fRand(double fMin, double fMax)
{
    double f = (double)rand() / RAND_MAX;
    return fMin + f * (fMax - fMin);
}
*/


/* Computes the distance between two nodes (1) and (2) */
static float dist(float x1, float y1, float z1, float x2, float y2, float z2)
{
	return sqrt(pow(x1 - x2, 2) + pow(y1 - y2, 2) + pow(z1 - z2, 2));
}

/* Computes the angle formed by the two lines (1-p) and (2-p) */
static float angle(float x1, float y1, float z1, 
	float x2, float y2, float z2, 
	float xp, float yp, float zp)
{
	float ux, uy, uz, vx, vy, vz;
	ux = x2-x1;
	uy = y2-y1;
	uz = z2-z1;
	vx = xp-x1;
	vy = yp-y1;
	vz = zp-z1;
	float a = ux*vx+uy*vy+uz*vz;
	float modU = sqrt(pow(ux, 2) + pow(uy, 2) + pow(uz, 2));
	float modV = sqrt(pow(vx, 2) + pow(vy, 2) + pow(vz, 2));
	float cosfi = a / (modU * modV);
	return acos(cosfi);
}

/* Computes the projected distance between nodes (p) and (2)
 * on the line (1-2) */
static float projDist(float x1, float y1, float z1, 
	float x2, float y2, float z2, 
	float xpa, float ypa, float zpa)
{
	float xp = xpa - x1;
	float yp = ypa - y1;
	float zp = zpa - z1;
	float a, b, c;
	a = (x2-x1) / dist(x1,y1,z1,x2,y2,z2);
	b = (y2-y1) / dist(x1,y1,z1,x2,y2,z2);
	c = (z2-z1) / dist(x1,y1,z1,x2,y2,z2);
	float xpp, ypp, zpp;
	xpp = x1+(a*xp+b*yp+c*zp)*a;
	ypp = y1+(a*xp+b*yp+c*zp)*b;
	zpp = z1+(a*xp+b*yp+c*zp)*c;
	//printf("(%f,%f,%f)\n", xpp,ypp,zpp);

	return dist(xpp, ypp, zpp, x2, y2, z2);
	return 1;
}

static float tg(float x1, float y1, float x2, float y2)
{
	float tg;
	tg = atan2(y2 - y1, x2 - x1);
 	if (tg < 0)
    	tg += 2*M_PI;
  	return tg;
}

static int orientation(float x1, float y1, float x2, float y2, float xp, float yp)
{
	float val;
    val = (y2 - y1) * (xp - x2) - (x2 - x1) * (yp - y2);
    //fprintf(stderr, "val = %f\n", val);
    if (abs(val) <= 0.001) return 0;  // colinear (prevent floating error)
    return (val > 0)? 1: 2; // clock or counterclock wise
}

static bool intersect(float x1, float y1, float x2, float y2, float x3, float y3, float x4, float y4)
{
	int o1 = orientation(x1, y1, x2, y2, x3, y3);
	int o2 = orientation(x1, y1, x2, y2, x4, y4);
	int o3 = orientation(x3, y3, x4, y4, x1, y1);
	int o4 = orientation(x3, y3, x4, y4, x2, y2);
	//fprintf(stderr, "o1 = %d, o2=%d, o3=%d, o4=%d\n", o1,o2,o3,o4);
	if((o1 != o2 && o3 != o4 && o3 != 0))
		return true;
	return false;
}

static bool insideBox(float xp, float yp, float zp, 
					float x1, float y1, float z1, 
					float x2, float y2, float z2, 
					float r)
{
	// coordinate x
	if(x1 <= x2) { // 1---2
		if(xp < x1 - r || xp > x2 + r) return false;
	} else { // 2---1
		if(xp < x2 - r || xp > x1 + r) return false;
	}

	// coordinate y
	if(y1 <= y2) { // 1--2
		if(yp < y1 - r || yp > y2 + r) return false;
	} else { // 2---1
		if(yp < y2 - r || yp > y1 + r) return false;
	}

	// coordinate z
	if(z1 <= z2) { // 1---2
		if(zp < z1 - r || zp > z2 + r) return false;
	} else { // 2---1
		if(zp < z2 - r || zp > z1 + r) return false;
	}
	return true;
}

// Returns 1 if the lines intersect, otherwise 0. In addition, if the lines 
// intersect the intersection point may be stored in the floats i_x and i_y.
static bool get_line_intersection(float p0_x, float p0_y, float p1_x, float p1_y, 
    float p2_x, float p2_y, float p3_x, float p3_y, float *i_x, float *i_y)
{
    float s1_x, s1_y, s2_x, s2_y;
    s1_x = p1_x - p0_x;     s1_y = p1_y - p0_y;
    s2_x = p3_x - p2_x;     s2_y = p3_y - p2_y;

    float s, t;
    s = (-s1_y * (p0_x - p2_x) + s1_x * (p0_y - p2_y)) / (-s2_x * s1_y + s1_x * s2_y);
    t = ( s2_x * (p0_y - p2_y) - s2_y * (p0_x - p2_x)) / (-s2_x * s1_y + s1_x * s2_y);

    if (s >= 0 && s <= 1 && t >= 0 && t <= 1)
    {
        // Collision detected
        if (i_x != NULL)
            *i_x = p0_x + (t * s1_x);
        if (i_y != NULL)
            *i_y = p0_y + (t * s1_y);
        return true;
    }

    return false; // No collision
}

static void solve( int n, float a[3][3], float b[3], float x[3]) {
  float s;
 
  for( int i = 0; i < n; i++) {
        s = 0;
        for( int j = 0; j < i; j++) {
            s = s + a[ i][ j] * x[ j];
        }
        x[ i] = ( b[ i] - s) / a[ i][ i];
   }
}

struct gpoint {
float x;
float y;
float z;
};


static gpoint ortoPoint(float x1, float y1, float z1,
				     float x2, float y2, float z2, 
				     float x3, float y3, float z3) {
	float a,b,c;
	a=(y2-y1)*(z3-z1)-(z2-z1)*(y3-y1); //cout<<"a: "<<a<<endl;
	b=(z2-z1)*(x3-x1)-(x2-x1)*(z3-z1); //cout<<"b: "<<b<<endl;
	c=(x2-x1)*(y3-y1)-(y2-y1)*(x3-x1); //cout<<"c: "<<c<<endl;

	float xp, yp, zp;
	xp = x1 + a / dist(x1, y1, z1, x3, y3, z3);
	yp = y1 + b / dist(x1, y1, z1, x3, y3, z3);
	zp = z1 + c / dist(x1, y1, z1, x3, y3, z3);
	gpoint ortoPoint;
	ortoPoint.x = xp;
	ortoPoint.y = yp;
	ortoPoint.z = zp;
	return ortoPoint;
}